//////////////////////////////////////////////////////////////////////
// headtracker.c
//
//////////////////////////////////////////////////////////////////////


// Standard includes
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <math.h>
#include <assert.h>
#include <string.h>
#include <memory.h>
#include <gl/gl.h>

// Our includes
#include <stdio.h>
#include "base.h"
#include "indycam.h"

// The minimum and maximum length of the ellipse's minor axis
#define MIN_OUTLINE_SZ   15
#define MAX_OUTLINE_SZ   25

// The maximum possible dimensions of the search area;
// used for statically allocating arrays
#define MAX_SEARCH_X     30
#define MAX_SEARCH_Y     30

// The threshold used to saturate the gradient values
static int thresh_absderiv = 85;

// Search range is +/- {msx, msy} pixels
static int msx = 8, msy = 8; 

// Threshold used in computing curvature
static int thresh_curv = 15;

// Skip szinc pixels to find the smaller and larger scale
static int szinc = 1;

// Structure for error functions
static struct {
	int max;      // maximum of error function
	int curv;     // "curvature" of error function
	int fn[2*MAX_SEARCH_Y+1][2*MAX_SEARCH_X+1]; 	// error function
} err[3];

// Structure for the outline's state
typedef struct {
     int x;         // x location
     int y;         // y location
     int sz;        // size
} state;

typedef struct {
     int perim_length;
     short *perim;
     int halfsize_x;
     int halfsize_y;
     unsigned char *mask;
} OUTLINE;

// Images
static ushort *gauss;
static uchar *img, *grad;
static uchar *imgRGBA;		// for displaying
BOOL displayOnlyImages = FALSE;
static int ncols = 160, nrows = 120;	// Must be divisible by 640,480

// Outlines
OUTLINE outlines[MAX_OUTLINE_SZ + 1];
static int conf; // confidence

// The current state (x,y,sz) of the ellipse
static state curr_state;

// The confidence is linearly scaled between these two values,
// to make the display more compelling.
static int min_conf = 60, max_conf = 90;

// Controls whether image or gradient is displayed
static BOOL displayGrad = FALSE;

//////////////////////////////////////////////////////////////////////
// error
// 
// Prints an error message and dies.
// 
// INPUTS
// 	exactly like printf
// 
// OUTPUTS
// 	none
//////////////////////////////////////////////////////////////////////

void error(char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	fprintf(stderr, "Error: ");
	vfprintf(stderr, fmt, args);
	fprintf(stderr, "\n");
	va_end(args);
	exit(1);
}


//////////////////////////////////////////////////////////////////////
// warning
// 
// Prints a warning message.
// 
// INPUTS
// 	exactly like printf
// 
// OUTPUTS
// 	none
//////////////////////////////////////////////////////////////////////

void warning(char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	fprintf(stderr, "Warning: ");
	vfprintf(stderr, fmt, args);
	fprintf(stderr, "\n");
	fflush(stderr);
	va_end(args);
}


//////////////////////////////////////////////////////////////////////
// smoothImage
// 
// Smooths an image by convolving with a 5x5 Gaussian, using
// two fixed one-dimensional kernels:  [1 4 6 4 1].  After
// the full convolution, the result is divided by 8, so that the
// maximum pixel value is 8192.
// 
// INPUTS
// imagebuf:  grey-scale image
// ncols, nrows:  size of image
// 
// OUTPUTS
// gaussbuf:  smoothed image
// 
// NOTES
// Sets the two-pixel out-of-bounds border to zero.
//////////////////////////////////////////////////////////////////////

void smoothImage(unsigned char *imagebuf, unsigned short *gaussbuf,
			  int ncols, int nrows)
{
	unsigned char *cptr;
	unsigned short *sptr;
	int i, j, indx;

	// Zero top border
	sptr = gaussbuf;
	for (i = 0 ; i < 2*ncols + 2 ; i++)  
		*sptr++ = 0;

	// Convolve middle with horizontal [1 4 6 4 1] mask
	cptr = imagebuf;
	sptr = gaussbuf + 2*ncols + 2;
	for (i = 2 ; i < (nrows-2) ; i++)  {
		for (j = 2 ; j < ncols - 2 ; j++)  {
			indx = i*ncols+j;
			*sptr++ = cptr[indx-2] + 4 * cptr[indx-1] + 6 * cptr[indx]
					+ 4 * cptr[indx+1] + cptr[indx+2];
 		}
		for (j = 1 ; j < 5 ; j++)
			*sptr++ = 0;
	}

	// Convolve middle with vertical [1 4 6 4 1]^T mask
	cptr = imagebuf;
	sptr = gaussbuf + 2*ncols + 2;
	for (i = 2 ; i < (nrows-2) ; i++)  {
		for (j = 2 ; j < ncols - 2 ; j++)  {
			indx = i*ncols+j;
			*sptr += cptr[indx-2*ncols] + 4 * cptr[indx-1*ncols] 
					+ 6 * cptr[indx]
					+ 4 * cptr[indx+1*ncols] + cptr[indx+2*ncols];
			*sptr++ /= 8;
 		}
		sptr += 4;
	}

	// Zero bottom border
	cptr = imagebuf + (nrows-2)*ncols + 2;
	sptr = gaussbuf + (nrows-2)*ncols + 2;
	for (i = (nrows-2)*ncols + 2 ; i < ncols*nrows ; i++)
		*sptr++ = 0;
}


//////////////////////////////////////////////////////////////////////
// computeGradient
// 
// Takes the absolute value of the derivative of an image by
// convolving the image with two one-dimensional filters ([-1 0 1]
// and [-1 0 1]^T), taking the absolute values, and summing.
//  
// INPUTS
// gauss:  smoothed image, as returned by smoothImage()
// ncols, nrows:  size of image
// 
// OUTPUTS
// grad:  absolute derivative
// 
// NOTES
// Sets the two-pixel out-of-bounds border to zero.
// Saturates the values above thresh_absderiv; that is, any value
// above thresh_absderiv becomes 255.
//////////////////////////////////////////////////////////////////////

void computeGradient(unsigned short *gauss, unsigned char *grad,
				 int ncols, int nrows)
{
	register unsigned short *ptrL, *ptrR, *ptrU, *ptrD;
	register unsigned char *ptro;
	register unsigned char *end;
	register int val, row;


	// Zero top border
	for (ptro = grad ; ptro < grad+3*ncols+3 ; *ptro++ = 0) ;

	end = grad + ncols*(nrows-3) - 3;
	for (ptrL = gauss+3*ncols+2, ptrR = gauss+3*ncols+4,
		 ptrU = gauss+2*ncols+3, ptrD = gauss+4*ncols+3 ;
		 ptro < end; )  {

		val = abs(*ptrL++ - *ptrR++) + abs(*ptrU++ - *ptrD++);
		if (val > thresh_absderiv)  val = 255;

		// Clear left and right three-pixel borders
		row = (ptro - grad) % ncols;
		if (row <= 2 || row >= ncols - 3)  val = 0;

		assert(ptro >= grad);
		assert(ptro < grad + ncols*nrows*sizeof(char));
		*ptro++ = val;
	}

	// Zero bottom border
	for ( ; ptro < grad+ncols*nrows ; *ptro++ = 0) ;
}


//////////////////////////////////////////////////////////////////////
// changeGradientThreshold
// getGradientThreshold
//
// Provides an interface for changing and getting the gradient 
// threshold.
//////////////////////////////////////////////////////////////////////

void changeGradientThreshold(int gt)
{
	if (gt < 0)  {
		warning("Gradient threshold is less than zero; setting to zero.");
		gt = 0;
	}
	else if (gt > 255)  {
		warning("Gradient threshold is greater than 255; setting to 255.");
		gt = 255;
	}
	thresh_absderiv = gt;
}

int getGradientThreshold(void)
{
	return thresh_absderiv;
}


//////////////////////////////////////////////////////////////////////
// makeEllipticalOutline
// 
// Creates an elliptical outline, used as a model for the head.  
// 
// INPUTS
// xrad, yrad:  length of the minor and major axes, respectively,
// 			 of the ellipse (in pixels).
// 
// NOTES
// xrad < yrad
//////////////////////////////////////////////////////////////////////

void makeEllipticalOutline(OUTLINE *outline, int xrad, int yrad)
{
	short *ptr;
	int xradsq = xrad * xrad;
	int yradsq = yrad * yrad;
	int xradbkpt = (int) (xradsq / sqrt(xradsq + yradsq));
	int yradbkpt = (int) (yradsq / sqrt(xradsq + yradsq));
	int x, y;
	int curx, cury;

	// count # of iterations in loops below
	outline->perim_length = 4 * xradbkpt + 4 * yradbkpt + 4; 
	outline->perim = (short *)
			malloc(2 * (outline->perim_length) * sizeof(short));
	ptr = outline->perim;
	*(ptr)++ = curx = 0;  // perimeter starts at top and proceeds clockwise
	*(ptr)++ = cury = - yrad;

	// from 0 to 45 degrees, measured cw from top
	for (x = 1; x <= xradbkpt ; x++)  {
		y = (int) (-yrad * sqrt(1 - ((float) x / xrad)*((float) x / xrad)));
		*(ptr)++ = x - curx;
		*(ptr)++ = y - cury;
		curx = x;
		cury = y;
	}
	
	// from 45 to 135 degrees (including right axis)
	for (y = - yradbkpt ; y <= yradbkpt ; y++)  {
		x = (int) (xrad * sqrt(1 - ((float) y / yrad)*((float) y / yrad)));
		*(ptr)++ = x - curx;
		*(ptr)++ = y - cury;
		curx = x;
		cury = y;
	}
	// from 135 to 225 degrees (including down axis)
	for (x = xradbkpt  ; x >= - xradbkpt ; x--)  {
		y = (int) (yrad * sqrt(1 - ((float) x / xrad)*((float) x / xrad)));
		*(ptr)++ = x - curx;
		*(ptr)++ = y - cury;
		curx = x;
		cury = y;
	}
	// from 225 to 315 degrees (including left axis)
	for (y = yradbkpt ; y >= - yradbkpt ; y--)  {
		x = (int) (-xrad * sqrt(1 - ((float) y / yrad)*((float) y / yrad)));
		*(ptr)++ = x - curx;
		*(ptr)++ = y - cury;
		curx = x;
		cury = y;
	}
	// from 315 to 360 degrees
	for (x = - xradbkpt ; x < 0 ; x++)  {
		y = (int) (-yrad * sqrt(1 - ((float) x / xrad)*((float) x / xrad)));
		*(ptr)++ = x - curx;
		*(ptr)++ = y - cury;
		curx = x;
		cury = y;
	}
	
	outline->halfsize_x = xrad;
	outline->halfsize_y = yrad;
}


//////////////////////////////////////////////////////////////////////
// computeCurvature
// 
// Given the 3 error functions, approximates their curvatures by
// the number of pixels that are far from the maximum.
// 
// OUTPUTS
// err:  structure in which results are saved.
//////////////////////////////////////////////////////////////////////

static void computeCurvature(void)
{
	int i, j, k;
	int far_pixels, maxx;

	for (k = 0 ; k < 3 ; k++)  {
		far_pixels = 0;
		maxx = err[k].max;

		for (j = 0 ; j < (2*msy+1) ; j++)
			for (i = 0 ; i < (2*msx+1) ; i++)
				far_pixels += (maxx - err[k].fn[j][i] > thresh_curv);
		err[k].curv = (int) (100 * far_pixels / 
						((2*msx+1) * (2*msy+1)));
	}
}


//////////////////////////////////////////////////////////////////////
// perimGrad
// 
// Integrates the gradient along the perimeter of an outline.
//////////////////////////////////////////////////////////////////////

static int perimGrad(OUTLINE *outline, unsigned char *grad, 
			int x_cen, int y_cen)
{
	int x, y;
	short *ptr;
	unsigned char *ptrgrad;
	int sum;
	int i;
	int incx, incy;

	ptr = outline->perim;
	x = x_cen + *ptr++;
	y = y_cen + *ptr++;
	ptrgrad = grad + (ncols * y) + x;

	sum = 0;
	for (i = 1 ; i < outline->perim_length ; i++)  {
		assert(ptrgrad >= grad 
			   && ptrgrad < (grad + ncols*nrows*sizeof(char)));
		sum += *ptrgrad;
		incx = *ptr++;
		incy = *ptr++;
		ptrgrad += (ncols * incy) + incx;
	}	

	return(sum);
}


//////////////////////////////////////////////////////////////////////
// searchForHead
//
// INPUTS
// s:  old state (x,y,sz)
// 
// OUTPUTS
// snew:  new state (x,y,sz)
//
// RETURNS
// Confidence that the search has successfully found the head.
//////////////////////////////////////////////////////////////////////

int searchForHead(state *s, state *snew)
{
	register int score, bestscore = 0;
	state beststate;
	int conf, num_of_functions;
	register int sz, x, y;
	register int perim;
	register int i, j, k;

	assert(msx <= MAX_SEARCH_X);
	assert(msy <= MAX_SEARCH_Y);

	// Initialize error stuff
	for (k = 0 ; k < 3 ; k++)  {
		err[k].max = -1;
		for (i = 0 ; i < (2*msx+1) ; i++)
			for (j = 0 ; j < (2*msy+1) ; j++)  {
				assert(k >= 0);
				assert(k <= 2);
				assert(j >= 0);
				assert(j < 2*MAX_SEARCH_Y+1);
				assert(i >= 0);
				assert(i < 2*MAX_SEARCH_X+1);
				err[k].fn[j][i] = 0;
			}
	}

	// Search for best ellipse state
	bestscore = 0;
	for (sz = max(s->sz - szinc, MIN_OUTLINE_SZ), 
	 	 k = (sz - s->sz + szinc) / szinc ; 
	     sz <= min(s->sz + szinc, MAX_OUTLINE_SZ) ; 
		 sz += szinc, k++)  {

		perim = outlines[sz].perim_length;
		for (y = max(s->y - msy, outlines[sz].halfsize_y), 
			 j = y - s->y + msy ;  
			 y <= min(s->y + msy, nrows-1-outlines[sz].halfsize_y) ;
			 y++, j++)  {

			for (x = max(s->x - msx, outlines[sz].halfsize_x), 
				 i = x - s->x + msx ; 
				 x <= min(s->x + msx, ncols-1-outlines[sz].halfsize_x) ; 
				 x++, i++)  {
				score = perimGrad(&outlines[sz], grad, x, y);
				score /= perim;

				assert(k >= 0);
				assert(k <= 2);
				assert(j >= 0);
				assert(j < 2*MAX_SEARCH_Y+1);
				assert(i >= 0);
				assert(i < 2*MAX_SEARCH_X+1);
				err[k].fn[j][i] = score;
				if (score > bestscore)  {
					beststate.x = x;
					beststate.y = y;
					beststate.sz = sz;
					bestscore = score;
				}
				if (score > err[k].max)
					err[k].max = score;
			}
		}
	}

	// Determine confidence by looking at "curvature" of error functions
	computeCurvature();
	conf = 0;
	num_of_functions = 0;
	for (i = 0 ; i < 3 ; i++)
		if (err[i].curv > 0)  {
			conf += err[i].curv;
			num_of_functions++;
		}
	if (num_of_functions == 0)  {  // This should never happen
		num_of_functions = 1;
	}
	conf /= num_of_functions;

	snew->x = beststate.x;
	snew->y = beststate.y;
	snew->sz = beststate.sz;
	return(conf);
}


//////////////////////////////////////////////////////////////////////
// changeSearchRange
// getSearchRangeX
// getSearchRangeY
// 
// Provides an interface for changing and getting msx and msy.
//////////////////////////////////////////////////////////////////////

void changeSearchRange(int search_x, int search_y)
{
	if (search_x > MAX_SEARCH_X || search_y > MAX_SEARCH_Y)
		error("Search range %2d x %2d is too big.  "
			 "Either decrease range or change limits "
			 "in findhead.c.\n", search_x, search_y);

	msx = search_x;
	msy = search_y;
}

int getSearchRangeX(void)
{
	return msx;
}

int getSearchRangeY(void)
{
	return msy;
}


//////////////////////////////////////////////////////////////////////
// changeCurvatureThreshold
// getCurvatureThreshold
// 
// Provides an interface for changing and getting the curvature 
// threshold.
//////////////////////////////////////////////////////////////////////

void changeCurvatureThreshold(int ct)
{
	if (ct < 0)  {
		warning("Curvature threshold should be nonnegative; changing to zero.");
		ct = 0;
	}
	thresh_curv = ct;
}

int getCurvatureThreshold(void)
{
	return thresh_curv;
}


//////////////////////////////////////////////////////////////////////
// processImage
//
// Computes the gradient, searches for the head location in the new 
// image, and updates the ellipse's state.
//////////////////////////////////////////////////////////////////////

void processImage(void)
{
	state new_state;
	static int xvel = 0, yvel = 0;	// velocities (for prediction)

	smoothImage(img, gauss, ncols, nrows);
	computeGradient(gauss, grad, ncols, nrows);

	conf = searchForHead(&curr_state, &new_state);
	xvel = new_state.x - curr_state.x;
	yvel = new_state.y - curr_state.y;
	curr_state.x = new_state.x;
	curr_state.y = new_state.y;
	curr_state.sz = new_state.sz;
}


//////////////////////////////////////////////////////////////////////
// displayAll
//////////////////////////////////////////////////////////////////////

void displayAll(void)
{
	// Translate from B/W to color
	{
	   uchar *pin, *pend, *dataptr;
	   ulong *pout;

	   dataptr = (displayGrad ? grad : img);
	   pend = dataptr + ncols*nrows;
	   for (pin = dataptr, pout = (ulong *) imgRGBA ; pin < pend ; )
		   *pout++ = *pin++ * 0x00010101;
	}

	// Display image
	lrectwrite(0, 0, ncols-1, nrows-1, (ulong *) imgRGBA);

	// Overlay ellipse
	{
	   short *ptrs;
	   short loc[2];
	   int sz, maxsz;
	   int x, y;
	   int i;

	   cpack(0xFF0000FF);
	   bgnpoint();
	   maxsz = min(curr_state.sz+1, MAX_OUTLINE_SZ);
	   for (sz = maxsz-1 ; sz <= maxsz ; sz++)  {
		   loc[0] = curr_state.x;		// x
		   loc[1] = curr_state.y;		// y
		   ptrs = outlines[sz].perim;
     	   for (i = 0 ; i < outlines[sz].perim_length ; i++)  {
          	   loc[0] += *ptrs++;
          	   loc[1] += *ptrs++;
          	   v2s(loc);
     	   }
	   }
	   endpoint();
	}

	// Draw confidence
	{
	   int scaled_conf;

	   if (conf < min_conf)  scaled_conf = 0;
	   else if (conf > max_conf)  scaled_conf = 100;
	   else scaled_conf = (conf - min_conf) / (max_conf - min_conf);
	   cpack(0xFF00FF00);
	   sboxf((Coord) (ncols-100)/2,					(Coord) nrows - 3,
		    (Coord) (ncols-100)/2 + scaled_conf,		(Coord) nrows - 1);
	   cpack(0xFF0000FF);
	   sboxf((Coord) (ncols-100)/2 + scaled_conf + 1,	(Coord) nrows - 3,
		 (Coord) ncols - (ncols-100)/2,			(Coord) nrows - 1);
	}
}


//////////////////////////////////////////////////////////////////////
// usage
//////////////////////////////////////////////////////////////////////

static void usage(char *command)
{
   fprintf(stderr, 
	"\nElliptical headtracker, using live images from the Indycam.\n"
	"usage:  %s [-h] [-grad] [-s range] [-gt thresh] [-ct thresh]\n"
	"        [-h] displays this help message\n"
	"        [-grad] displays the gradient of the image\n"
	"        [-s range] sets the search range in both x and y (default: %d & %d)\n"
	"        [-gt thresh] sets the saturation threshold for the gradient;\n"
	"                     Any value above 'thresh' becomes 255 (default: %d).\n"
	"        [-ct thresh] sets the distance threshold for counting far\n"
	"                     pixels, in the curvature computation (default: %d).\n\n",
	command, getSearchRangeX(), getSearchRangeY(), 
	getGradientThreshold(), getCurvatureThreshold());
   exit(1);
}


//////////////////////////////////////////////////////////////////////
// main
//
// Main routine
//////////////////////////////////////////////////////////////////////

void main(int argc, char *argv[])
{
     long win;
     char *ptrin, *ptrout;
	int i, j;
	Coord ll, rr, bb, tt;

     // Parse command line
     for (i = 1 ; i < argc ; i++)  {
          if (strcmp(argv[i], "-h") == 0)  {
               usage(argv[0]);
          }
          else if (strcmp(argv[i], "-grad") == 0)  {
               displayGrad = TRUE;
          }
          else if (strcmp(argv[i], "-s") == 0)  {
			i++;
			if (i == argc)  error("Missing argument.");
			changeSearchRange(atoi(argv[i]), atoi(argv[i]));
          }
          else if (strcmp(argv[i], "-gt") == 0)  {
			i++;
			if (i == argc)  error("Missing argument.");
			changeGradientThreshold(atoi(argv[i]));
          }
          else if (strcmp(argv[i], "-ct") == 0)  {
			i++;
			if (i == argc)  error("Missing argument.");
			changeCurvatureThreshold(atoi(argv[i]));
          }
		else warning("Unknown argument.");
     }

	// Initialize camera
     openCamera(ncols, nrows);
     setupCamera(BW);
     // printf("xsize %d ysize %d\n", ncols, nrows);

	// Initialize elliptical outlines
	for (i = MIN_OUTLINE_SZ ; i <= MAX_OUTLINE_SZ ; i++)  {
		makeEllipticalOutline(&outlines[i], i, (int) (1.2 * i));
	}
	curr_state.x = ncols/2;
	curr_state.y = nrows/2;
	curr_state.sz = MAX_OUTLINE_SZ;

	// Allocate memory
	gauss = (ushort *) malloc(ncols * nrows * sizeof(short));
	grad = (uchar *) malloc(ncols * nrows * sizeof(char));
	imgRGBA = (uchar *) malloc(4 * ncols * nrows * sizeof(char));

     // Set up and open a GL window to display the data
     foreground();
     prefsize(ncols, nrows);
     win = winopen("Headtracker");
     RGBmode();
     pixmode(PM_TTOB, 1);
     gconfig();
	ortho2(0.0, (Coord) (ncols-1), (Coord) (nrows-1), 0.0); // flip y-axis
	clear();


	// Begin main loop
     printf("Press Ctrl-C to exit.\n");
     beginImageTransfer();
     for(;;) {
          img = waitForNextImage();
		processImage();
//		fprintf(stderr, "(%3d, %3d) %3d\n", xloc, yloc, sizeout);
		displayAll();
          freeImage();
     }

	// Clean up
     endImageTransfer();
     closeCamera();
}
