//////////////////////////////////////////////////////////////////////
// indycam.c
// 
// This file contains a (simple) driver for the SGI IndyCam.
//////////////////////////////////////////////////////////////////////

#include <unistd.h> 	// sginap
#include <dmedia/vl.h> 
#include <dmedia/vl_vino.h>   /* used for shutter speed */

#include "indycam.h"

// global variables
static VLServer svr;
static VLPath path;
static VLNode src, drn;
static VLControlValue val;
static VLBuffer buffer;


//////////////////////////////////////////////////////////////////////
// error_exit
//////////////////////////////////////////////////////////////////////

void error_exit(void)
{
    vlPerror("IndyCam: ");
    exit(1);
}


//////////////////////////////////////////////////////////////////////
// openCamera
// 
// INPUTS
// xsize, ysize:  desired image size.  These must divide into the
// default image size (which is usually 640x480).
//////////////////////////////////////////////////////////////////////

void openCamera(int xsize, int ysize)
{
    int width, height; 		// default image size

    // Connect to the daemon
    if (!(svr = vlOpenVideo(""))) 
        error_exit(); 

    // Set up a drain node in memory
    drn = vlGetNode(svr, VL_DRN, VL_MEM, VL_ANY);

    // Set up a source node on any video source
    src = vlGetNode(svr, VL_SRC, VL_VIDEO, VL_ANY);      

    // Create a path using the first device that will support it
    path = vlCreatePath(svr, VL_ANY, src, drn);

    // Set up the hardware for and define the usage of the path
    if ((vlSetupPaths(svr, (VLPathList)&path, 1, VL_SHARE, VL_LOCK)) < 0)
        error_exit();

    // Make sure that desired image size is OK
    vlGetControl(svr, path, drn, VL_SIZE, &val);
    width = val.xyVal.x;
    height = val.xyVal.y;
    if (width % xsize != 0 || height % ysize != 0)  {
        fprintf(stderr, "Error:  The default image size (%3d,%3d) is not"
                        "divisible by the desired size (%3d,%3d).\n",
                         width, height, xsize, ysize);
        error_exit();
    }
    if (width / xsize != height / ysize)  {
        fprintf(stderr, "Error:  The default image size (%3d,%3d) is"
                        "divisible by the desired size (%3d,%3d)"
                        "only by subsampling x different from y.\n",
                         width, height, xsize, ysize);
        error_exit();
    }
    if (width / xsize <= 0)  {
        fprintf(stderr, "Error:  The default image size (%3d,%3d) is"
                        "smaller than the desired size (%3d,%3d).\n",
                         width, height, xsize, ysize);
        error_exit();
    }

    // Set the zoom factor
    val.fractVal.numerator = 1;
    val.fractVal.denominator = (width / xsize);
    if (vlSetControl(svr, path, drn, VL_ZOOM, &val))
        error_exit();

    // Ensure that image size was set properly
    vlGetControl(svr, path, drn, VL_SIZE, &val);
    if (val.xyVal.x != xsize || val.xyVal.y != ysize)  {
        fprintf(stderr, "Error:  Unable to set image size.\n");
        error_exit();
    }
}


//////////////////////////////////////////////////////////////////////
// setupCamera
//
// Currently, just sets the packing mode, but could be augmented
// to initialize shutter speed, etc.
//////////////////////////////////////////////////////////////////////

void setupCamera(imageType imtype)
{
  val.intVal = (imtype == COLOR ? VL_PACKING_RGB_8 : VL_PACKING_Y_8_P);
  vlSetControl(svr, path, drn, VL_PACKING, &val); 
}


//////////////////////////////////////////////////////////////////////
// beginImageTransfer
//////////////////////////////////////////////////////////////////////

void beginImageTransfer()
{
    // Create and register a buffer for 3 frames --
    // This allows 99 ms for processing while still
    // allowing access to the most recent frame
    const int num_frames = 1;

    buffer = vlCreateBuffer(svr, path, drn, num_frames);
    if (buffer == NULL)
        error_exit();   
    vlRegisterBuffer(svr, path, drn, buffer);
    
	// Receive interleaved (interlaced) video
	val.intVal = VL_CAPTURE_INTERLEAVED;
	vlSetControl(svr, path, drn, VL_CAP_TYPE, &val);

    // Begin the data transfer
    if (vlBeginTransfer(svr, path, 0, NULL))
        error_exit();
}

//////////////////////////////////////////////////////////////////////
// waitForNextImage
//
// Waits until a new image is available, then returns a pointer
// to the image data.
//////////////////////////////////////////////////////////////////////

char *waitForNextImage()
{
	VLInfoPtr info;
     char *dataPtr;

	   // Wait until the next image is available
        do {
            sginap(1);          // wait a tick
            info = vlGetNextValid(svr, buffer);
        } while (!info);
    
        // Get a pointer to the frame
        dataPtr = vlGetActiveRegion(svr, buffer, info);
	return(dataPtr);
}
                

//////////////////////////////////////////////////////////////////////
// freeImage
//
// Unlock the ring buffer to allow the image to be overwritten.
// Call this function when the image is no longer needed.
//////////////////////////////////////////////////////////////////////

void freeImage()
{
        vlPutFree(svr, buffer);
}


//////////////////////////////////////////////////////////////////////
// endImageTransfer
//
// Ends the data transfer.
//////////////////////////////////////////////////////////////////////

void endImageTransfer()
{
    vlEndTransfer(svr, path);
}


//////////////////////////////////////////////////////////////////////
// closeCamera
//////////////////////////////////////////////////////////////////////

void closeCamera()
{
    // Cleanup before exiting
    vlDeregisterBuffer(svr, path, drn, buffer);
    vlDestroyBuffer(svr, buffer);
    vlDestroyPath(svr, path);
    vlCloseVideo(svr);
}

